function Int = Trapzc_adapt(a, b, fun, N, n)
%Trapzc_adapt: Automatic Composite Trapezoidal rule.
%   Int = Trapzc_adapt(a, b, N, fun, n) calculates the numerical approximation to
%   the definite integral of a function on equidistant intervals using
%   the composite trapezoidal quadrature rule. The number of intervals is
%   halved at every iteration until convergence is achieved or N > n.
%   Inputs:
%       a:   Starting point of the integration interval 
%       b:   Ending point of the integration interval (b > a)
%       fun: Integrand function (either inline, anonymous or defined in a
%            m-file)
%       N:   Initial value for the number of discretization intervals
%       (INTEGER) (default: 3)
% .     n:   Max number of iterations (default: 1e4)
%   Output:
% .     Int: Value of the definite integral




%If max number of iterations is not provided, set a default value
if ~exist('n', 'var')
    n=1e4;
end
if ~exist('N', 'var')
    N=3;
end


int0 = Trapzc(a, b, N, fun);
N = 2*N;
int1 = Trapzc(a, b, N, fun);

while abs(int0 - int1)>1e-6 && N<n
    N = 2*N;
    int0 = int1;
    int1 = Trapzc(a, b, N, fun);  
end

%Broadcast message if the number of iterations has exceeded the provided
%value
if N >= n
    warning(sprintf('Number of iterations exceeded %G . Convergence not achieved', N));
end


Int = int1;


end