import numpy as np
import matplotlib.pyplot as plt
from typing import Callable
from scipy.integrate import quad

def midpointc(a: float, b: float, N: int, fun: Callable[[np.ndarray], np.ndarray]) -> float:
    """
    Function that calculates the numerical approximation to
    the definite integral of a function on N equidistant intervals using
    the composite midpoint quadrature rule.

    :param a: lower integration bound
    :param b: upper integration bound
    :param N: No of integration intervals
    :param fun: Function to be inegrated

    :return : the value of the integral
    """

    #Define stepsize
    step = (b - a) / N
    #Generate vector of midpoints
    midpoints = np.linspace(a + step / 2, b - step/2, N)

    return step * np.sum(fun(midpoints))

def trapInt(a: float, b: float, N: int, fun: Callable[[np.ndarray], np.ndarray]) -> float:
    """
        Function that calculates the numerical approximation to
        the definite integral of a function on N equidistant intervals using
        the composite Trapezoidal quadrature rule.

        :param a: lower integration bound
        :param b: upper integration bound
        :param N: No of integration intervals
        :param fun: Function to be inegrated

        :return : the value of the integral
    """
    # Define stepsize
    step = (b - a) / N
    # Generate vector of nodes =
    nodes = np.linspace(a, b, N + 1)
    resInt = fun(nodes)
    resInt[[0, -1]] = resInt[[0, -1]] / 2

    return step * np.sum(resInt)

def simpsonc(a: float, b: float, N: int, fun: Callable[[np.ndarray], np.ndarray]) -> float:
    """
        Function that calculates the numerical approximation to
        the definite integral of a function on N equidistant intervals using
        the composite Simpson quadrature rule.

        :param a: lower integration bound
        :param b: upper integration bound
        :param N: No of integration intervals
        :param fun: Function to be inegrated

        :return : the value of the integral
    """
    # Define stepsize
    step = (b - a) / N
    # Generate vector of nodes
    nodes = np.linspace(a, b, N + 1)
    fTemp = fun(nodes)
    fTemp[1:-1] = 2 * fTemp[1:-1]
    resInt = step * np.sum(fTemp) / 6

    #Now midpoints
    nodes = np.linspace(a + step / 2, b - step/2, N)
    fTemp = fun(nodes)

    return resInt + 2 * step * np.sum(fTemp) / 3





if __name__ == "__main__":
    # Define parameters for nice plots
    plt.rcParams['font.family'] = 'Helvetica'
    plt.rcParams['xtick.major.width'] = 1.9
    plt.rcParams['xtick.major.size'] = 4.5
    plt.rcParams['xtick.minor.width'] = 1.5
    plt.rcParams['xtick.minor.size'] = 3
    plt.rcParams['ytick.major.width'] = 1.9
    plt.rcParams['ytick.major.size'] = 4.5
    plt.rcParams['ytick.minor.width'] = 1.5
    plt.rcParams['ytick.minor.size'] = 3
    plt.rcParams['font.size'] = 20
    plt.rcParams['axes.linewidth'] = 2.2
    plt.rcParams['figure.figsize'] = [14 / 2.54, 14 / 2.54]

    """

    1 Numerical Integration

    """
    # Function parameters
    cInf = 0
    c0 = 1
    D = 10
    N = 10
    zeta = np.linspace(0, 3, 100)
    funcInt = lambda s: np.exp(-s ** 2)

    # Integrate with scipy.integrate.quad
    c = [c0 + (cInf - c0) * quad(funcInt, 0, s)[0] for s in zeta]
    cTrapz = [c0 + (cInf - c0) * trapInt(0, s, N, funcInt) for s in zeta]

    # Plot concentration profile
    fig, ax = plt.subplots()
    ax.plot(zeta, c, lw=2, color="r")
    ax.plot(zeta, cTrapz, lw=2, color="b")

    ax.set_xlim(zeta.min(), zeta.max())
    ax.set_ylim(0, 1)
    ax.set_xlabel(r"$\zeta = \frac{z}{\sqrt{4Dt}}$")
    ax.set_ylabel("Concentration")

    plt.tight_layout()

    """

    2 Numerical Integration

    """

    #Paramters and integrand definition
    a = -2
    b = 2
    #N = np.round(np.logspace(1, 5, 100))
    N = np.logspace(1, 5, 100, dtype=np.intc)
    step = (b - a) / N
    fun = lambda x: np.exp(x)

    #Analytical solution
    intAnalytical = fun(b) - fun(a)
    #Numerical solution
    intNumerical = {"midpoint": np.asanyarray([midpointc(a, b, n, fun) for n in N]),
                    "trapezoidal": np.asanyarray([trapInt(a, b, n, fun) for n in N]),
                    "simpson": np.asanyarray([simpsonc(a, b, n, fun) for n in N])}

    errRel = {key: np.abs(val - intAnalytical) / intAnalytical for (key, val) in intNumerical.items()}

    #Order of accuracy
    for key in errRel.keys():
        if key == "simpson":
            #Reduce fitting range
            orderAcc = np.polyfit(np.log(step[np.where(step > 5e-3)]),
                          np.log(errRel[key][np.where(step > 5e-3)]),
                          deg=1)


        else:
            orderAcc = np.polyfit(np.log(step),
                          np.log(errRel[key]),
                          deg=1)

        print(f"Order of accuracy for {key} method: {orderAcc[0]:.0f}")



    fig, ax = plt.subplots()
    for label in errRel.keys():
        ax.loglog(step, errRel[label], lw=2, marker='o', label=label)

    ax.set_xlim(step.min(), step.max())
    ax.legend(frameon=False, loc="upper left")
    ax.grid(which="both", lw=0.5, alpha=0.8)
    ax.set_xlabel("Step size")
    ax.set_ylabel("Relative error")
    plt.tight_layout(pad=0.4)


    plt.show()


